import pkg_resources
from typing import Optional

import pandas as pd
from rpy2.robjects import pandas2ri
from rpy2.rinterface import RRuntimeError, NULL
from rpy2.robjects.packages import (
    importr,
    SignatureTranslatedAnonymousPackage as STAPkg
)

from .spec import ArmaGarchSpec
from .exceptions import ArmaGarchException


class ArmaGarchProcess:
    """
     A time series model with means modelled as an ARMA process and
     volatilities modelled as a GARCH process

     This class wraps a limited subset of the R package rugarch.
     See ArmaGarchSpec for details on how to set the spec
    """

    r_lib = None

    def __new__(cls, *args, **kwargs):
        """ import the R library functions """
        if not cls.r_lib:
            importr('rugarch')
            src = pkg_resources.resource_filename('pygarch', 'lib/fit.R')
            with open(src, 'r') as script:
                cls.r_lib = STAPkg(script.read(), 'r_lib')
        return super().__new__(cls)

    def __init__(self, series: pd.Series,
                 arma_order: tuple = (1, 1),
                 garch_order: tuple = (1, 1)):
        if series.empty:
            raise ValueError('series is empty!')

        self.spec = ArmaGarchSpec(arma_order, garch_order)

        # internal
        self._index = series.index
        self._init_price = series.iloc[-1]
        self._model_converged = False

        # r objects
        pandas2ri.activate()
        self._r_series = pandas2ri.py2ri(series)
        self._r_model = None

    def __str__(self):
        return (f"<ARMA{self.spec.arma_order}-GARCH{self.spec.garch_order} "
                f"Process at {hex(id(self))}>")

    def fit(self) -> bool:
        """ fit an ARMA-GARCH model to the series """
        try:
            res = ArmaGarchProcess.r_lib.fit_model(self._r_series, self.spec.as_r_obj())
            self._r_model, converged = res
            self._model_converged = converged[0]
        except RRuntimeError:
            self._model_converged = False
        return self._model_converged

    @property
    def converged(self) -> bool:
        return self._model_converged

    def simulate(self,
                 len_sim: int,
                 num_sims: int,
                 external_regressors: Optional[pd.DataFrame] = None
                 ) -> pd.DataFrame:
        """
         Generate simulations from the ARMA-GARCH model starting from the
         period following the end date of the fitting-period
        """
        if (self._r_model is None) or (not self.converged):
            raise ArmaGarchException(
                "cannot simulate: model has not been fit or has not converged"
            )

        extr_fit = self.spec.external_regressors  # fitted on
        extr_sim = external_regressors            # to simulate on
        if isinstance(extr_sim, pd.Series):
            extr_sim = extr_sim.to_frame()

        if not extr_fit.empty and (extr_sim is None):
            raise ValueError(
                "cannot simulate: must supply simulations of external regressors"
            )

        elif extr_fit.empty and (extr_sim is not None):
            raise ValueError(
                "cannot simulate: model was not fit with external_regressors"
            )

        elif not extr_fit.empty and (extr_sim is not None):
            if extr_sim.shape != (len_sim, extr_sim.shape[1]):
                raise ValueError(
                    f"external_regressors is wrong dimension. It must be "
                    f"of size (len_sim, N)=({len_sim}, {extr_sim.shape[1]})"
                )

        if extr_sim is None:
            extr_sim = NULL

        sims = ArmaGarchProcess.r_lib.simulate_model(
            self._r_model, len_sim, num_sims, extr_sim
        )
        # TODO: update this when we update pandas
        return pd.DataFrame(pandas2ri.ri2py_vector(sims))

    @property
    def aic(self) -> float:
        """ return the Akaike Information Criteria """
        if self.converged:
            return ArmaGarchProcess.r_lib.calc_aic(self._r_model, self._r_series)[0]
        else:
            raise ArmaGarchException("model has not been fit or has not converged")

    @property
    def summary(self) -> str:
        """ return the summary of the model fit """
        if self.converged:
            rvec = ArmaGarchProcess.r_lib.get_model_summary(self._r_model)
            return '\n'.join(tuple(rvec))
        else:
            raise ArmaGarchException("model has not been fit or has not converged")

    @property
    def residuals(self) -> Optional[pd.Series]:
        """ return the residuals from the mean model """
        if self.converged:
            ret = ArmaGarchProcess.r_lib.get_residuals(self._r_model)
            return pd.Series(pandas2ri.ri2py(ret).ravel(), index=self._index)
        else:
            return None

    @property
    def fitted_values(self) -> Optional[pd.Series]:
        """ return the fitted values from the mean model """
        if self.converged:
            ret = ArmaGarchProcess.r_lib.get_fitted_values(self._r_model)
            return pd.Series(pandas2ri.ri2py(ret).ravel(), index=self._index)
        else:
            return None

    @property
    def volatilities(self) -> Optional[pd.Series]:
        """ return the fitted volatilities from the GARCH component on the model """
        if self.converged:
            ret = ArmaGarchProcess.r_lib.get_volatilities(self._r_model)
            return pd.Series(pandas2ri.ri2py(ret).ravel(), index=self._index)
        else:
            return None



from setuptools import setup, find_packages

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='pygarch',
    version="0.0.1",
    author="Oliver Chambers",
    long_description=long_description,
    url='https://gitlab.corp.vesparum.com/oliver.chambers/pygarch',
    packages=find_packages(),
    install_requires=[
        "pandas",
        "rpy2"
    ],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Topic :: Utilities',
    ],
)



import pkg_resources
from typing import Optional

import pandas as pd
from rpy2 import rinterface
from rpy2.robjects import pandas2ri
from rpy2.robjects.packages import (
    importr,
    SignatureTranslatedAnonymousPackage as STAPkg
)


class ArmaGarchSpec:
    """
     A limited wrapper to the R objects rugarch::ugarchspec

     Usage:
       arma_order -> tuple [default=(1,1)] :
           the number of autoregressive and moving average terms respectively
           in the ARMA process (of the means)
       garch_order -> tuple [default=(1,1)] :
           the number of volatility terms and (ARCH) error terms respectively
           in the GARCH process (of the volatilities)
       external_regressors -> pandas.DataFrame (optional) :
           additional external regressors that predict the mean
       dist -> str ["norm", "std"] [default="norm"] :
           the error distribution. Gaussian (norm) or Student-t (std)
       dof -> int (optional) :
           the degrees of the freedom in the Student-t distribution. If not
           specified then it is fit
       fixed_ar -> tuple (optional) :
           a tuple of coefficients for the autoregressive terms of the ARMA
           model.
       fixed_ma -> tuple (optional) :
           a tuple of coefficients for the moving-average terms of the ARMA
           model.
       model_mean -> bool [default=True] :
           whether to model the mean in the ARMA process (like the intercept
           in a linear regression
    """

    r_lib = None

    def __new__(cls, *args, **kwargs):
        if not cls.r_lib:
            importr('rugarch')
            lib_path = pkg_resources.resource_filename('pygarch', 'lib/spec.R')
            with open(lib_path, 'r') as script:
                cls.r_lib = STAPkg(script.read(), 'r_lib')
        return super().__new__(cls)

    def __init__(self,
                 arma_order: tuple = (1, 1),
                 garch_order: tuple = (1, 1),
                 external_regressors: Optional[pd.DataFrame] = None,
                 dist: str = "norm",
                 dof: Optional[int] = None,
                 fixed_ar: Optional[tuple] = None,
                 fixed_ma: Optional[tuple] = None,
                 model_mean: bool = True):

        pandas2ri.activate()

        self.arma_order = arma_order
        self.garch_order = garch_order
        self.external_regressors = external_regressors
        self.dist = dist
        self.dof = dof
        self.fixed_ar = fixed_ar
        self.fixed_ma = fixed_ma
        self.model_mean = model_mean

    def as_r_obj(self):
        return ArmaGarchSpec.r_lib.create_spec(
            self._arma_order, self._garch_order,
            self._external_regressors, self._dist, self._dof,
            self._fixed_ar, self._fixed_ma, self.model_mean
        )

    @property
    def arma_order(self) -> tuple:
        return tuple(self._arma_order)

    @arma_order.setter
    def arma_order(self, order: tuple) -> None:
        if len(order) != 2:
            raise ValueError(
                'arma_order must be a tuple of length 2 that specifies the '
                'number of autoregressive and moving-average terms'
            )
        self._arma_order = rinterface.IntSexpVector(order)

    @property
    def garch_order(self) -> tuple:
        return tuple(self._garch_order)

    @garch_order.setter
    def garch_order(self, order: tuple) -> None:
        if len(order) != 2:
            raise ValueError(
                'garch_order must be a tuple of length 2 that specifies the '
                'number of volatility terms and error terms respectively'
            )
        self._garch_order = rinterface.IntSexpVector(order)

    @property
    def external_regressors(self) -> pd.DataFrame:
        if not self._external_regressors:
            return pd.DataFrame()
        else:
            # TODO: update this when we update pandas
            return pd.DataFrame(pandas2ri.ri2py_vector(self._external_regressors))

    @external_regressors.setter
    def external_regressors(self, frame: Optional[pd.DataFrame]) -> None:
        if frame is None:
            self._external_regressors = rinterface.NULL
        else:
            self._external_regressors = pandas2ri.py2ri(frame)

    @property
    def dist(self) -> str:
        return self._dist

    @dist.setter
    def dist(self, dist: str) -> None:
        if dist not in ("norm", "std"):
            raise ValueError("distribution must be 'norm' for Gaussian error or 'std' "
                             "for Student-t errors")
        self._dist = dist

    @property
    def dof(self) -> Optional[int]:
        return int(self._dof) if self._dof else None

    @dof.setter
    def dof(self, dof: Optional[int]) -> None:
        if dof is not None:
            if self.dist != "std":
                raise ValueError("degrees of freedom (dof) can only be set with "
                                 "the Student-t error distribution")
            elif dof <= 0:
                raise ValueError("degrees of freedom (dof) must be positive")

        self._dof = dof if (dof is not None) else rinterface.NULL

    @property
    def fixed_ar(self) -> Optional[tuple]:
        if not self._fixed_ar:
            return None
        return tuple(self._fixed_ar)

    @fixed_ar.setter
    def fixed_ar(self, coeffs: Optional[tuple]) -> None:
        if not coeffs:
            self._fixed_ar = rinterface.NULL
        elif len(coeffs) > self.arma_order[0]:
            raise ValueError(
                f"the fixed autoregressive coefficients cannot be longer than "
                f"the arma order, {self.arma_order[0]}"
            )
        else:
            self._fixed_ar = rinterface.IntSexpVector(coeffs)

    @property
    def fixed_ma(self) -> Optional[tuple]:
        if not self._fixed_ma:
            return None
        return tuple(self._fixed_ma)

    @fixed_ma.setter
    def fixed_ma(self, coeffs: Optional[tuple]) -> None:
        if not coeffs:
            self._fixed_ma = rinterface.NULL
        elif len(coeffs) > self.arma_order[1]:
            raise ValueError(
                f"the fixed autoregressive coefficients cannot be longer than "
                f"the arma order, {self.arma_order[1]}"
            )
        else:
            self._fixed_ma = rinterface.IntSexpVector(coeffs)
